import { RegionType, ServiceType } from "../../interface";
import { waitResponse } from "@ygkit/request";
import { ApiCredentials } from "../../interface";
import { CdnDeployInputs } from "./interface";
import { pascalCaseProperty, removeUndefined, sleep } from "../../utils";
import { Client } from "tencentcloud-sdk-nodejs/tencentcloud/services/cdn/v20180606/cdn_client";
import { BriefDomain } from "tencentcloud-sdk-nodejs/tencentcloud/services/cdn/v20180606/cdn_models";
import { formatter } from "./formatter";
import { ApiTypeError } from "../../error";
import { Capi } from "@tencent-sdk/capi";

export const ONE_SECOND = 1000;
export const TIMEOUT = 5 * 60 * ONE_SECOND;

export default class Cdn {
    credentials: ApiCredentials;
	client: Client;
	capi: Capi;

    constructor(credentials: ApiCredentials = {} as any) {
        this.credentials = credentials;

        this.client = new Client({
            credential: {
                secretId: credentials.SecretId,
                secretKey: credentials.SecretKey,
                token: credentials.Token,
            },
            region: "ap-guangzhou",
		});
		
		this.capi = new Capi({
			ServiceType: ServiceType.cdn,
			Region: RegionType["ap-guangzhou"],
			SecretId: this.credentials.SecretId,
			SecretKey: this.credentials.SecretKey,
			Token: this.credentials.Token,
		});
    }

    /**
     * purgeCdnCache 用于批量提交目录刷新，根据域名的加速区域进行对应区域的刷新。
     * 内部调用 PurgeCdnUrls 接口
     */
    async purgeCdnUrls(urls: string[], flushType = "flush") {
        console.log(`Purging CDN caches, it will work in 5 minutes...`);
        try {
            await this.client.PurgePathCache({
                Paths: urls,
                FlushType: flushType,
            });
        } catch (e) {
            // no op
        }
    }

    /**
     * pushCdnUrls 用于将指定 URL 资源列表加载至 CDN 节点，支持指定加速区域预热。
     * 内部调用 PushUrlsCache 接口
     * @param urls url 资源
     * @param userAgent 指定预热请求回源时 HTTP 请求的 User-Agent 头部，默认为 flush
     * @param area 预热生效区域
     */
    // FIXME: userAgent 默认应为 TencentCDN
    async pushCdnUrls(urls: string[], userAgent = "flush", area = "mainland") {
        console.log(`Pushing CDN caches...`);
        try {
            await this.client.PushUrlsCache({
                Urls: urls,
                Area: area,
                UserAgent: userAgent,
            });
        } catch (e) {
            // no op
        }
    }

    /**
     * 根据域名查询 CDN 信息
     * 内部调用 DescribeDomains
     */
    async getCdnByDomain(domain: string) {
        const { Domains } = await this.client.DescribeDomains({
            Filters: [{ Name: "domain", Value: [domain] }],
        });

        if (Domains && Domains.length) {
            return Domains[0];
        }
        return undefined;
    }

    /**
     * 下线指定域名的 CDN 服务
     * @param domain 域名
     */
    async offlineCdnDomain(domain: string) {
        const { Status } = await this.getCdnByDomain(domain);
        if (Status === "online") {
            // disable first
            await this.client.StopCdnDomain({ Domain: domain });
        } else if (Status === "processing") {
            throw new Error(
                `Status is not operational for ${domain}, status: ${Status}`
            );
        }
    }

	/** 启动用户的 CDN 服务 */
    async openCdnService() {
        try {
            await this.capi.request(
                {
                    Action: "OpenCdnService",
                    PayTypeMainland: "flux",
                    PayTypeOverseas: "flux",
                },
                {
                    isV3: false,
                    debug: false,
                    RequestClient: "TencentCloud - ServerlessComponentCdn",
                    host: `${ServiceType.cdn}.tencentcloudapi.com`,
                    path: "/",
                }
            );
        } catch (e) {
            if (e.code !== "ResourceInUse.CdnUserExists") {
                throw e;
            }
        }
    }

    /** 部署 CDN */
    async deploy(inputs: CdnDeployInputs) {
        await this.openCdnService();
        const { oldState } = inputs;

        delete inputs.oldState;

        /** 仅刷新，不部署 */
        if (inputs.onlyRefresh === true) {
            const domainExist = await this.getCdnByDomain(inputs.domain);
            // refresh cdn urls
            if (domainExist && inputs.refreshCdn && inputs.refreshCdn.urls) {
                await this.purgeCdnUrls(
                    inputs.refreshCdn.urls,
                    inputs.refreshCdn.flushType
                );
            }

            return {
                resourceId: domainExist.ResourceId,
                https: !!inputs.https,
                domain: inputs.domain,
                origins: inputs.origin && inputs.origin.origins,
                cname: `${inputs.domain}.cdn.dnsv1.com`,
                refreshUrls: inputs.refreshCdn.urls,
            };
        }

        const cdnInputs: CdnDeployInputs = removeUndefined({
            ...inputs,
            origin: formatter.originInputs(inputs.origin),
        });

        const sourceInputs = JSON.parse(JSON.stringify(cdnInputs));

        const outputs = {
            https: !!inputs.https,
            domain: inputs.domain,
            origins: cdnInputs.origin.origins,
            cname: `${inputs.domain}.cdn.dnsv1.com`,
            inputCache: JSON.stringify(inputs),
            resourceId: "",
        };

        let cdnInfo = await this.getCdnByDomain(inputs.domain);

        // pass state for cache check
        const { inputCache } = oldState;
        if (inputCache && inputCache === JSON.stringify(sourceInputs)) {
            console.log(
                `No configuration changes for CDN domain ${inputs.domain}`
            );
            outputs.resourceId = cdnInfo.ResourceId;
        } else {
            await this.createOrUpdateCdn(cdnInputs);
        }

        return outputs;
    }

    async createOrUpdateCdn(inputs: CdnDeployInputs, cdnInfo?: BriefDomain) {
        const outputs = {
            resourceId: "",
        };

        if (cdnInfo) {
            cdnInfo = await this.getCdnByDomain(inputs.domain);
		}
		
		/** 如果 CDN 已关闭，先删除再创建 */
        if (cdnInfo && cdnInfo.Status === "offline") {
            console.log(`The CDN domain ${inputs.domain} is offline.`);
            console.log(`Recreating CDN domain ${inputs.domain}`);
            await this.client.DeleteCdnDomain({ Domain: inputs.domain });
            cdnInfo = null;
		}
		
		/** 如果 CDN 存在且未关闭，更新 */
        if (cdnInfo) {
            // update
            console.log(`The CDN domain ${inputs.domain} has existed.`);
            console.log("Updating...");
            // TODO: when update, VIP user can not set ServiceType parameter, need CDN api optimize
            if (
                inputs.serviceType &&
                String(inputs.serviceType) !== cdnInfo.ServiceType
            ) {
                inputs.serviceType = cdnInfo.ServiceType;
            }
            await this.client.UpdateDomainConfig(pascalCaseProperty(inputs));
            outputs.resourceId = cdnInfo.ResourceId;
        } else {
		
		/** 创建 CDN */
            console.log(`Adding CDN domain ${inputs.domain}...`);
            try {
                // if not config ServiceType, default to web
                inputs.serviceType = inputs.serviceType ?? "web";
                await this.client.AddCdnDomain(pascalCaseProperty(inputs));
            } catch (e) {
                if (e.code === "ResourceNotFound.CdnUserNotExists") {
                    console.log(
                        `Please goto https://console.cloud.tencent.com/cdn open CDN service.`
                    );
                }
                throw e;
            }
            await sleep(1000);
            const detail = await this.getCdnByDomain(inputs.domain);

            outputs.resourceId = detail && detail.ResourceId;
        }
        console.log(
            "Waiting for CDN deploy success, it maybe cost 5 minutes...."
        );

        // 是否等待 CDN 部署完毕
        // When set syncFlow false, just continue, do not wait for online status.
        if (inputs.async === false) {
            await waitResponse({
                callback: async () => this.getCdnByDomain(inputs.domain),
                targetProp: "Status",
                targetResponse: "online",
                timeout: TIMEOUT,
            });

            // FIXME: purge first, push next?
            // push cdn urls
            if (inputs.pushCdn?.urls) {
                await this.pushCdnUrls(
                    inputs.pushCdn.urls,
                    inputs.pushCdn.area,
                    inputs.pushCdn.userAgent
                );
            }

            // refresh cdn urls
            if (inputs.refreshCdn) {
                await this.purgeCdnUrls(inputs.refreshCdn.urls);
            }
        }
        console.log(`CDN deploy success to domain: ${inputs.domain}`);
    }

    /** 删除 CDN */
    async remove(inputs: { domain: string }) {
        const { domain } = inputs;
        if (!domain) {
            throw new ApiTypeError(`PARAMETER_CDN`, "domain is required");
        }

        // need circle for deleting, after domain status is 6, then we can delete it
        console.log(`Start removing CDN for ${domain}`);
        const detail = await this.getCdnByDomain(domain);
        if (!detail) {
            console.log(`CDN domain ${domain} not exist`);
            return {};
        }

        const { Status } = detail;

        if (Status === "online") {
            // disable first
            await this.client.StopCdnDomain({ Domain: domain });
        } else if (Status === "processing") {
            console.log(`Status is not operational for ${domain}`);
            return {};
        }
        console.log(`Waiting for offline ${domain}...`);
        await waitResponse({
            callback: async () => this.getCdnByDomain(domain),
            targetProp: "Status",
            targetResponse: "offline",
            timeout: TIMEOUT,
        });
        console.log(`Removing CDN for ${domain}`);
        await this.client.DeleteCdnDomain({ Domain: domain });
        console.log(`Removed CDN for ${domain}.`);
        return {};
    }
}
